// file      : xsd/cxx/parser/parser-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (ename (l));
          SemanticGraph::Type& t (l.argumented ().type ());
          String type_name (ename (t));

          //@@ What's to come is a nasty hack.
          //
          String item_name;

          if (t.is_a<SemanticGraph::Fundamental::IdRef> ())
          {
            item_name = L"IDREF" == name ? L"IDREF_" : L"IDREF";
          }
          else if (t.is_a<SemanticGraph::Fundamental::IdRefs> ())
          {
            item_name = L"IDREFS" == name ? L"IDREFS_" : L"IDREFS";
          }
          else
            item_name = type_name == name ? type_name + L"_" : type_name;


          os << "template < typename _xsd_" << name << ", "
             << "typename _xsd_" << item_name << " >" << endl
             << "struct " << name
             << ": ::xsd::cxx::parser::list< " <<
            "_xsd_" << name << ", "
            "_xsd_" << item_name << ", " <<
            char_type << ", " <<
            parser_template << " >"
             << "{"
             << "// You would normally override the pre, item," << endl
             << "// and post functions in your implementation."
             << "};";
        }
      };


      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          String const& name (ename (u));

          os << "template < typename _xsd_" << name << " >" << endl
             << "struct " << name <<
            ": " << parser_template << "< _xsd_" << name <<
            ", " << char_type << " >"
             << "{"
             << "// You would normally override the pre, _characters," << endl
             << "// and post functions in your implementation."
             << "};";
        }
      };


      //
      //
      struct BaseInheritance: Base
      {
        BaseInheritance (Context& c)
            : Context (c), Base (c)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          os << ", ";
          Base::traverse (c);
        }
      };


      //
      //
      struct ParserHook : Traversal::Member,
                          protected virtual Context
      {
        ParserHook (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name);

          os << "virtual void" << endl
             << name << " (" << type << " const&);"
             << endl;
        }
      };


      //
      //
      struct ParserAccessor : Traversal::Member,
                              protected virtual Context
      {
        ParserAccessor (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name);

          os << "void" << endl
             << name << "_parser (" << parser_template << "< " <<
            type << ", " << char_type << " >&);"
             << endl;
        }
      };


      //
      //
      struct ParserMember : Traversal::Member,
                            protected virtual Context
      {
        ParserMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name);
          String inst (L"_xsd_" + name + L"_");

          os << parser_template << "< " << type << ", " <<
            char_type << " >* " << inst << ";";
        }
      };


      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              base_inheritance_ (c),
              template_param_decl_list_ (c),
              parser_hook_ (c),
              parser_member_ (c),
              parser_accessor_ (c)
        {
          inherits_ >> base_inheritance_;

          names_parser_hook_ >> parser_hook_;
          names_parser_member_ >> parser_member_;
          names_parser_accessor_ >> parser_accessor_;
        }

        virtual Void
        traverse (Type& c)
        {
          String const& name (ename (c));
          String type (L"_xsd_" + name);
          Boolean has_members (has<Traversal::Element> (c) ||
                               has<Traversal::Attribute> (c));

          os << "template < ";

          {
            template_param_decl_list_.traverse (c);
          }

          os << " >" << endl;

          os << "struct " << name
             << ": " << parser_template << "< " << type << ", " <<
            char_type << " >";

          {
            enter_scope (name);
            inherits<Complex> (c, inherits_, 0, &Complex::nl, &Complex::nl);
            leave_scope ();
          }

          os << "{";

          if (has_members)
          {
            os << "// Parser hooks. Override them in your implementation." << endl
               << "//" << endl
               << endl;

            enter_scope (name);
            names (c, names_parser_hook_);
            leave_scope ();
          }

          os << "using " << parser_template << "< " << type << ", " <<
            char_type << " >::pre;"
             << "using " << parser_template << "< " << type << ", " <<
            char_type << " >::post;"
             << endl
             << endl;

          //
          //
          if (has_members)
          {
            os << "// Parser construction API." << endl
               << "//" << endl
               << endl;

            enter_scope (name);
            names (c, names_parser_accessor_);
            leave_scope ();

            os << "void" << endl
               << "parsers (";

            {
              // Automatically enters the scope.
              //
              ParserParamDecl decl (*this, false);
              decl.traverse (c);
            }

            os << ");"
               << endl
               << endl;
          }


          os << "// Implementation." << endl
             << "//" << endl
             << "protected:" << endl
             << endl;


          // _start_element
          //
          os << "virtual void" << endl
             << "_start_element (" <<
            string_type << " const& ns, " <<
            string_type << " const& n);"
             << endl;

          // end_element
          //
          os << "virtual void" << endl
             << "_end_element (" <<
            string_type << " const& ns, " <<
            string_type << " const& n);"
             << endl;


          // attribute
          //
          os << "virtual void" << endl
             << "_attribute (" <<
            string_type << " const& ns, " <<
            string_type << " const& n, " <<
            string_type << " const& v);"
             << endl;


          if (has_members)
          {
            os << "private:" << endl
               << endl;

            enter_scope (name);
            names (c, names_parser_member_);
            leave_scope ();
          }

          os << "};";
        }

        virtual Void
        nl (Type& c)
        {
          os << endl;
        }

      private:
        BaseInheritance base_inheritance_;
        Traversal::Inherits inherits_;

        //
        //
        TemplateParamDeclList template_param_decl_list_;

        //
        //
        ParserHook parser_hook_;
        Traversal::Names names_parser_hook_;

        //
        //
        ParserMember parser_member_;
        Traversal::Names names_parser_member_;

        //
        //
        ParserAccessor parser_accessor_;
        Traversal::Names names_parser_accessor_;
      };


      //
      //
      struct Includes : Traversal::Imports,
                        Traversal::Includes,
                        protected virtual Context
      {
        Includes (Context& c, String const& expr)
            : Context (c), expr_ (expr)
        {
        }

        virtual Void
        traverse (SemanticGraph::Imports& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Includes& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Path const& path)
        {
          using BackendElements::Regex::perl_s;

          // Note the use of the portable representation of the path.
          //
          String const& name (perl_s (String (path.string ()), expr_));

          Char op (include_with_brackets ? '<' : '"');
          Char cl (include_with_brackets ? '>' : '"');

          os << "#include " << op << include_prefix << name << cl << endl
             << endl;
        }

      private:
        String expr_;
      };

      struct FundamentalNamespace : Namespace,
                                    protected virtual Context
      {
        FundamentalNamespace (Context& c)
            : Context (c),
              Namespace (c)
        {
        }

        void
        traverse (Type& ns)
        {
          pre (ns);

          os << "// Generic parser." << endl
             << "//" << endl;

          String g (L"XSD_CXX_PARSER_" + xml_parser + L"_" + char_type +
                    L"_PARSER_IN_" + ns_name (ns));

          std::transform (g.begin (), g.end(), g.begin (), upcase);
          g = escape (g); // Make it a C++ id.

          os << "#ifndef " << g << endl
             << "#define " << g << endl
             << endl
             << "template <typename X>" << endl
             << "struct parser: ::xsd::cxx::parser::" << xml_parser <<
            "::traverser< X, " << char_type << " >" << endl
             << "{"
             << "};"
             << "#endif" << endl
             << endl
             << endl;

          os << "// Parsers for XML Schema built-in types." << endl
             << "//" << endl
             << endl;

          os << "// 8-bit" << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::byte< " <<
            char_type << ", " << parser_template << " > byte;"
             << "typedef ::xsd::cxx::parser::unsigned_byte< " <<
            char_type << ", " << parser_template << " > unsigned_byte;"
             << endl;


          os << "// 16-bit" << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::short_< " <<
            char_type << ", " << parser_template << " > short_;"
             << "typedef ::xsd::cxx::parser::unsigned_short< " <<
            char_type << ", " << parser_template << " > unsigned_short;"
             << endl;


          os << "// 32-bit" << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::int_< " <<
            char_type << ", " << parser_template << " > int_;"
             << "typedef ::xsd::cxx::parser::unsigned_int< " <<
            char_type << ", " << parser_template << " > unsigned_int;"
             << endl;


          os << "// 64-bit" << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::long_< " <<
            char_type << ", " << parser_template << " > long_;"
             << "typedef ::xsd::cxx::parser::unsigned_long< " <<
            char_type << ", " << parser_template << " > unsigned_long;"
             << endl;

          os << "// Boolean." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::boolean< " <<
            char_type << ", " << parser_template << " > boolean;"
             << endl;

          os << "// Floating-point types." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::float_< " <<
            char_type << ", " << parser_template << " > float_;"
             << "typedef ::xsd::cxx::parser::double_< " <<
            char_type << ", " << parser_template << " > double_;"
             << "typedef ::xsd::cxx::parser::decimal< " <<
            char_type << ", " << parser_template << " > decimal;"
             << endl;

          os << "// String." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::string< " <<
            char_type << ", " << parser_template << " > string;"
             << endl
             << endl;


          os << "// Exceptions. See libxsd/xsd/cxx/parser/exceptions.hxx " <<
            "for details." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::exception< " <<
            char_type << " > exception;"
             << "typedef ::xsd::cxx::parser::parsing< " <<
            char_type << " > parsing;"
             << "typedef ::xsd::cxx::parser::unexpected_element< " <<
            char_type << " > unexpected_element;"
             << endl
             << endl;

          if (xml_parser == L"xerces")
          {
            os << "// Parsing flags. See " <<
              "libxsd/xsd/cxx/parser/elements.hxx for details." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::xerces::flags flags;"
               << endl
               << endl;

            os << "// Parsing properties. See " <<
              "libxsd/xsd/cxx/parser/elements.hxx for details." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::xerces::properties< " <<
              char_type << " > properties;"
               << endl
               << endl;
          }

          if (xml_parser == L"expat")
          {
            os << "// Parsing Context." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::expat::context< " <<
              char_type << " > context;"
               << endl
               << endl;

            os << "// Simple auto pointer for Expat's XML_Parser object." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::expat::parser_auto_ptr " <<
              "parser_auto_ptr;"
               << endl
               << endl;
          }

          post (ns);
        }
      };
    }

    Void
    generate_parser_header (Context& ctx,
                            SemanticGraph::Schema& tu,
                            String const& expr)
    {
      String c (ctx.char_type);

      // At the moment we only support UTF-8 expat.
      //
      if (ctx.xml_parser == L"expat")
      {
        ctx.os << "#ifdef XML_UNICODE" << endl
               << "#error UTF-16 expat (XML_UNICODE defined) is not " <<
          "yet supported" << endl
               << "#endif" << endl
               << endl;
      }

      //
      //
      if (c == L"char")
      {
        ctx.os << "#ifndef XSD_USE_CHAR" << endl
               << "#define XSD_USE_CHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_PARSER_USE_CHAR" << endl
               << "#define XSD_CXX_PARSER_USE_CHAR" << endl
               << "#endif" << endl
               << endl;
      }
      else if (c == L"wchar_t")
      {
        ctx.os << "#ifndef XSD_USE_WCHAR" << endl
               << "#define XSD_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_PARSER_USE_WCHAR" << endl
               << "#define XSD_CXX_PARSER_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;
      }

      ctx.os << "#include <xsd/cxx/parser/exceptions.hxx>" << endl
             << "#include <xsd/cxx/parser/" << ctx.xml_parser <<
        "/elements.hxx>" << endl
             << "#include <xsd/cxx/parser/types.hxx>" << endl
             << endl;

      // Emit fundamental types.
      //
      {
        Traversal::Schema schema, xsd;
        Traversal::Implies implies;
        Traversal::Names names;
        FundamentalNamespace ns (ctx);

        schema >> implies >> xsd >> names >> ns;

        schema.dispatch (tu);
      }

      Traversal::Schema schema;

      Traversal::Sources sources;
      Includes includes (ctx, expr);
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> includes;
      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);

      names >> list;
      names >> union_;
      names >> complex;

      schema.dispatch (tu);
    }
  }
}

