// file      : xsd/cxx/tree/elements.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/elements.hxx>

namespace CXX
{
  namespace Tree
  {
    // GenerateDefautCtor
    //
    GenerateDefaultCtor::
    GenerateDefaultCtor (Context& c, Boolean& generate)
        : Context (c), generate_ (generate)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Type& t)
    {
      generate_ = true;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Enumeration& e)
    {
      generate_ = true;
    }

    Void GenerateDefaultCtor::
    traverse (SemanticGraph::Member& m)
    {
      if (!skip (m) && min (m) == 1 && max (m) == 1)
        generate_ = true;
    }


    // GenerateFromBaseCtor
    //
    GenerateFromBaseCtor::
    GenerateFromBaseCtor (Context& c, Boolean& generate)
        : traverser_ (c, generate)
    {
      inherits_ >> traverser_;
    }

    Void GenerateFromBaseCtor::
    traverse (SemanticGraph::Complex& c)
    {
      inherits (c, inherits_);
    }

    GenerateFromBaseCtor::Traverser::
    Traverser (Context& c, Boolean& generate)
        : Context (c), generate_ (generate)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void GenerateFromBaseCtor::Traverser::
    traverse (SemanticGraph::Member& m)
    {
      if (!skip (m) && min (m) == 1 && max (m) == 1)
        generate_ = true;
    }


    // FromBaseCtorArg
    //
    FromBaseCtorArg::
    FromBaseCtorArg (Context& c, Boolean arg)
        : Context (c), arg_ (arg)
    {
    }

    Void FromBaseCtorArg::
    traverse (SemanticGraph::Member& m)
    {
      if (skip (m)) return;

      // Note that we are not going to include attributes with
      // default values here. Instead we are going to default-
      // initialize them.
      //
      if (min (m) == 1 && max (m) == 1)
      {
        String const& name (ename (m));

        os << "," << endl
           << iscope_name (name, "type") << " const&";

        if (arg_)
          os << " _xsd_" << name;
      }
    }

    //
    //
    CtorArgs::
    CtorArgs (Context& c, String const& name, Boolean arg)
        : Context (c), name_ (name), arg_ (arg), first_ (true)
    {
      *this >> inherits_ >> *this;
      *this >> names_ >> *this;
    }

    Void CtorArgs::
    traverse (SemanticGraph::Type& t)
    {
      os << comma () << name_ << "::member_ const& " <<
        (arg_ ? name_ : String ("/* base */"));
    }

    Void CtorArgs::
    traverse (SemanticGraph::Enumeration& e)
    {
      os << comma () << name_ << "::member_ const& " <<
        (arg_ ? name_ : String ("/* base */"));
    }

    Void CtorArgs::
    traverse (SemanticGraph::Member& m)
    {
      if (skip (m)) return;

      // Note that we are not going to include attributes with
      // default values here. Instead we are going to default-
      // initialize them.
      //
      if (min (m) == 1 && max (m) == 1)
      {
        String const& name (ename (m));

        os << comma () << iscope_name (name, "type") << " const&";

        if (arg_)
          os << " _xsd_" << name;
      }
    }

    String CtorArgs::
    comma ()
    {
      Boolean tmp (first_);
      first_ = false;
      return tmp ? "" : ",\n";
    }


    // GlobalElementBase
    //
    Boolean GlobalElementBase::
    generate_p (SemanticGraph::Element& e)
    {
      return e.substitutes_p () || doc_root_p (e);
    }

    Boolean GlobalElementBase::
    doc_root_p (SemanticGraph::Element& e)
    {
      if (!options.value<CLI::root_element_first> () &&
          !options.value<CLI::root_element_last> () &&
          !options.value<CLI::root_element_all> () &&
          !options.value<CLI::root_element_none> () &&
          options.value<CLI::root_element> ().empty ())
        return true; // By default treat them all.

      if (options.value<CLI::root_element_none> ())
        return false;

      if (options.value<CLI::root_element_all> ())
        return true;

      if (options.value<CLI::root_element_first> () &&
          e.context ().count ("first") != 0)
        return true;

      if (options.value<CLI::root_element_last> () &&
          e.context ().count ("last") != 0)
        return true;

      typedef Cult::Containers::Vector<NarrowString> Names;
      Names const& names (options.value<CLI::root_element> ());

      // Hopefully nobody will specify more than a handful of names ;-).
      //
      for (Names::ConstIterator i (names.begin ()); i != names.end (); ++i)
      {
        String name (*i);

        if (e.name () == name)
          return true;
      }

      return false;
    }

    // Namespace
    //
    Namespace::
    Namespace (Context& c,
               Boolean set_scope,
               UnsignedLong first,
               UnsignedLong last)
        : CXX::Namespace (c, set_scope),
          GlobalElementBase (c),
          ctx_ (c), first_ (first), last_ (last), count_ (0)
    {
    }

    Void Namespace::
    traverse (Type& ns)
    {
      using SemanticGraph::Element;

      if (first_ > last_)
        CXX::Namespace::traverse (ns);
      else
      {
        Boolean opened (false);

        for (Type::NamesIterator i (ns.names_begin ());
             i != ns.names_end (); ++i)
        {
          SemanticGraph::Nameable& n (i->named ());

          if (n.is_a<SemanticGraph::Type> () ||
              (n.is_a<Element> () && generate_p (dynamic_cast<Element&> (n))))
          {
            if (count_ >= first_ && count_ <= last_)
            {
              if (!opened)
              {
                opened = true;
                pre (ns);
              }

              edge_traverser ().dispatch (*i);
            }

            ++count_;
          }
        }

        if (opened)
          post (ns);
      }
    }
  }
}
